#include <fcntl.h>
#include <stdint.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <libgen.h>
#include <json-c/arraylist.h>
#include <json-c/json_object_iterator.h>
#include <json-c/json_util.h>
#include <json-c/json_object.h>
#include <json-c/json_tokener.h>
#include "util.h"
#include "client.h"
#include "wm.h"

static unsigned int node_id = 0;

TreeNode* wm_treenode_new(NodeType type, TreeNode *parent)
{
    DEBUG_PRINT("%s\n", __func__);
    TreeNode *node = calloc(sizeof(TreeNode), 1);
    assert(node);

    node->parent = parent;
    node->type = type;
    node->children = wm_nodearray_new();

    node_id++;
    node->id = node_id;
    node->client = NULL;

    return node;
}

void wm_treenode_free(TreeNode *node)
{
    wm_nodearray_clear(node->children);
    wm_nodearray_free(node->children);
    free(node);
}

bool wm_treenode_is_empty(TreeNode *node)
{
    DEBUG_PRINT("%s\n", __func__);
    return (node->children->size == 0);
}

void wm_treenode_split_space(TreeNode *node, Rect *ret1, Rect *ret2)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(node);
    assert(ret1);
    assert(ret2);

    switch (node->type) {
    case NODE_VERTICAL:
        ret1->w = node->pos.w / 2;
        ret1->h = node->pos.h;
        ret2->w = node->pos.w / 2;
        ret2->h = node->pos.h;

        ret1->x = node->pos.x;
        ret1->y = node->pos.y;

        ret2->x = node->pos.x + ret1->w;
        ret2->y = node->pos.y;
        return;
    case NODE_HORIZONTAL:
        ret1->w = node->pos.w;
        ret1->h = node->pos.h / 2;
        ret2->w = node->pos.w;
        ret2->h = node->pos.h / 2;

        ret1->x = node->pos.x;
        ret1->y = node->pos.y;

        ret2->x = node->pos.x;
        ret2->y = node->pos.y + ret1->h;
        return;
    default:
        fprintf(stderr, "wm: unhandled treenode type %d in %s, exiting.\n",
                node->type, __func__);
        abort();
    }
}

void wm_treenode_recalculate_space(TreeNode *node)
{
    assert(node);

    assert(node->children->size == 2);

    wm_treenode_split_space(node, &node->children->nodes[0]->pos,
                            &node->children->nodes[1]->pos);
}

void wm_treenode_swap(TreeNode *root, TreeNode *node1, TreeNode* node2)
{
    assert(root);
    assert(node1);
    assert(node2);

    assert(node1->type == NODE_CLIENT);
    assert(node2->type == NODE_CLIENT);

    Client *tmp = node1->client;
    node1->client = node2->client;
    node2->client = tmp;
}

void wm_treenode_add_child(TreeNode *node, TreeNode *child)
{
    assert(node);
    assert(child);

    wm_nodearray_push(node->children, child);
    child->parent = node;
}

void wm_node_type_to_str(NodeType type, char *buf, size_t bufsize)
{
    switch (type) {
    case NODE_CLIENT:
        strncpy(buf, "NODE_CLIENT", bufsize);
        return;
    case NODE_VERTICAL:
        strncpy(buf, "NODE_VERTICAL", bufsize);
        return;
    case NODE_HORIZONTAL:
        strncpy(buf, "NODE_HORIZONTAL", bufsize);
        return;
    case NODE_TAB:
        strncpy(buf, "NODE_TAB", bufsize);
        return;
    case NODE_COUNT:
        return;
    }
}

NodeType wm_node_type_from_str(char *buf, size_t bufsize)
{
    char *types[4] = {0};
    types[NODE_CLIENT] = "NODE_CLIENT";
    types[NODE_VERTICAL] = "NODE_VERTICAL";
    types[NODE_HORIZONTAL] = "NODE_HORIZONTAL";
    types[NODE_TAB] = "NODE_TAB";

    for (size_t i = 0; i < 4; i++) {
        if (strncmp(buf, types[i], bufsize) == 0)
            return (NodeType)i;
    }

    fprintf(stderr, "wm: invalid node type string: %s. exiting\n", buf);
    exit(1);
}

char* wm_treenode_to_str(TreeNode *node)
{
    assert(node);
    char type[128] = {0};
    wm_node_type_to_str(node->type, type, sizeof(type));

    json_object *jobj = json_object_new_object();

    json_object_object_add(jobj, "type", json_object_new_string(type));
    json_object_object_add(jobj, "parent", json_object_new_uint64((uint64_t)node->parent));
    json_object_object_add(jobj, "children", json_object_new_array());

    for (size_t i = 0; i < node->children->size; i++) {
        char *child_str = wm_treenode_to_str(node->children->nodes[i]);
        json_object_array_add(json_object_object_get(jobj, "children"),
            json_tokener_parse(child_str));
        free(child_str);
    }

    json_object *pos = json_object_new_object();
    json_object_object_add(pos, "x", json_object_new_int(node->pos.x));
    json_object_object_add(pos, "y", json_object_new_int(node->pos.y));
    json_object_object_add(pos, "w", json_object_new_int(node->pos.w));
    json_object_object_add(pos, "h", json_object_new_int(node->pos.h));
    json_object_object_add(jobj, "pos", pos);
    if (node->type == NODE_CLIENT && node->client) {
        json_object *client = json_object_new_object();
        json_object_object_add(client, "address", json_object_new_uint64((unsigned long)node->client));
        json_object_object_add(client, "name", json_object_new_string(node->client->name));
        json_object_object_add(client, "is_pid_set", json_object_new_boolean(node->client->is_pid_set));
        json_object_object_add(client, "pid", json_object_new_int(node->client->pid));
        json_object_object_add(jobj, "client", client);
    }
    json_object_object_add(jobj, "id", json_object_new_uint64((uint64_t)node->id));

    char *ret = strdup(json_object_to_json_string_ext(jobj, JSON_C_TO_STRING_PRETTY));

    json_object_put(jobj);

    return ret;
}

UIntArray* wm_nonempty_workspaces_to_strptrarray(Wm *wm)
{
    _Static_assert(sizeof(uint64_t) == sizeof(uint64_t*), "static assert failed");
    UIntArray *ret = wm_uintarray_new();

    for (size_t i = 0; i < wm->selected_monitor->wscount; i++) {
        TreeNode *node = wm->selected_monitor->workspaces[i].tree;
        if (!wm_treenode_is_empty(node) || (node->type == NODE_CLIENT && node->client)) {
            WmWorkspaceToStrRet *ws_str_with_index = malloc(sizeof(WmWorkspaceToStrRet));
            assert(ws_str_with_index);
            *ws_str_with_index = (WmWorkspaceToStrRet) {
                .ws_index = i,
                .str = wm_treenode_to_str(wm->selected_monitor->workspaces[i].tree)
            };
            wm_uintarray_push(ret, (uint64_t)ws_str_with_index);
        }
    }

    DEBUG_PRINT("%s returning %ld\n", __func__, ret->size);

    return ret;
}

json_object* wm_state_to_json_object(Wm *wm)
{
    json_object *ret = json_object_new_object();

    UIntArray *ws_str = wm_nonempty_workspaces_to_strptrarray(wm);
    char str[128] = {0};

    for (size_t i = 0; i < ws_str->size; i++) {
        WmWorkspaceToStrRet *ws_with_index = (WmWorkspaceToStrRet*)
            wm_uintarray_at(ws_str, i);

        snprintf(str, sizeof(str), "%ld", ws_with_index->ws_index);
        json_object_object_add(ret, str, json_tokener_parse(ws_with_index->str));

        free(ws_with_index->str);
        free(ws_with_index);
    }
    wm_uintarray_free(ws_str);

    return ret;
}

void wm_log_state(Wm *wm, const char *prefixstr, const char* logfile)
{
    RETURN_IF_NULL(prefixstr);
    RETURN_IF_NULL(logfile);

    DEBUG_PRINT("saving state to file: %s\n", logfile);

    char prefix[WM_PREFIX_LEN + 32] = {0};
    snprintf(prefix, sizeof(prefix), "%lu:%s", time(NULL), prefixstr);

    json_object *log_entry_obj = wm_state_to_json_object(wm);

    int fd = -1;
    struct json_object *jobj = NULL;

    int access_ret = access(logfile, F_OK);
    if (access_ret != 0) {
        DEBUG_PRINT("logfile does not exist\n");
        fd = creat(logfile, 0644);

        if (fd < 0) {
            perror("creat");
            goto ret;
        }

        jobj = json_object_new_object();
    } else {
        DEBUG_PRINT("logfile exists\n");

        struct stat st;
        if (stat(logfile, &st) < 0) {
            perror("stat");
            goto ret;
        }

        fd = open(logfile, O_RDWR);
        if (fd < 0) {
            perror("open");
            goto ret;
        }

        if (st.st_size == 0) {
            DEBUG_PRINT("logfile is empty");
            jobj = json_object_new_object();
        } else {
            jobj = json_object_from_fd(fd);
            if (!jobj) {
                const char *err = json_util_get_last_err();
                fprintf(stderr, "wm: could not read json: %s\n", err);
                DEBUG_PRINT("wm: could not read json: %s\n", err);
                // TODO maybe free
                goto ret;
            }
        }
    }

    json_object_object_add(jobj, prefix, log_entry_obj);

    int ret = close(fd);
    if (ret < 0) {
        perror("close");
        goto ret;
    }

    fd = open(logfile, O_TRUNC | O_WRONLY);

    if (ret < 0) {
        perror("open");
        goto ret;
    }

    json_object_to_fd(fd, jobj, JSON_C_TO_STRING_PRETTY);

    ret = close(fd);
    if (ret < 0) {
        perror("close");
        goto ret;
    }

    // TODO
    wm_logentries_free(wm->log_entries);
    wm->log_entries = wm_read_log(logfile);

ret:
    json_object_put(jobj);
}

TreeNode* wm_treenode_lmd(TreeNode *node)
{
    assert(node);

    if (node->children->size == 0) return node;

    return wm_treenode_lmd(node->children->nodes[0]);
}

NodeArray* wm_treenode_all_lmds(TreeNode *node)
{
    assert(node);

    NodeArray *postorder = wm_postorder_traversal(node);
    NodeArray *ret = wm_nodearray_new();

    for (size_t i = 0; i < postorder->size; i++) {
        TreeNode *lmd = wm_treenode_lmd(postorder->nodes[i]);

        wm_nodearray_push(ret, lmd);
    }

    wm_nodearray_free(postorder);
    return ret;
}

UIntArray* wm_treenode_all_lmds_index(TreeNode *node)
{
    assert(node);

    NodeArray *postorder = wm_postorder_traversal(node);
    UIntArray *ret = wm_uintarray_new();

    for (size_t i = 0; i < postorder->size; i++) {
        TreeNode *lmd = wm_treenode_lmd(postorder->nodes[i]);

        uintptr_t index = UINT_MAX;
        for (index = 0; index < postorder->size; index++) {
            if (postorder->nodes[index]->id == lmd->id)
                break;
        }

        assert(index != UINT_MAX);
        wm_uintarray_push(ret, index);
    }

    wm_nodearray_free(postorder);
    return ret;
}

static void traversal(TreeNode *node, NodeArray *arr)
{
    if (!node) return;

    for (size_t i = 0; i < node->children->size; i++)
        traversal(node->children->nodes[i], arr);

    wm_nodearray_push(arr, node);
}

NodeArray* wm_postorder_traversal(TreeNode *tree)
{
    NodeArray *ret = wm_nodearray_new();

    traversal(tree, ret);

    return ret;
}

bool wm_is_treenode_keyroot(TreeNode *node, NodeArray *postorder)
{
    assert(node);
    assert(postorder);
    if (!node->parent) return true;

    TreeNode *node_lmd = wm_treenode_lmd(node);
    size_t node_index = SIZE_MAX;

    for (size_t i = 0; i < postorder->size; i++) {
        if (node->id == postorder->nodes[i]->id) {
            node_index = i;
            break;
        }
    }

    assert(node_index != SIZE_MAX);

    for (size_t i = node_index + 1; i < postorder->size; i++) {
        if (node_lmd->id == wm_treenode_lmd(postorder->nodes[i])->id)
            return false;
    }

    return true;
}

NodeArray* wm_treenode_all_keyroots(TreeNode *node)
{
    assert(node);

    NodeArray* ret = wm_nodearray_new();
    NodeArray* postorder = wm_postorder_traversal(node);

    for (size_t i = 0; i < postorder->size; i++) {
        TreeNode *node = postorder->nodes[i];
        if (wm_is_treenode_keyroot(node, postorder))
            wm_nodearray_push(ret, node);
    }

    wm_nodearray_free(postorder);
    return ret;
}

UIntArray* wm_treenode_all_keyroots_index(TreeNode *node)
{
    assert(node);

    NodeArray *postorder = wm_postorder_traversal(node);
    UIntArray *ret = wm_uintarray_new();

    for (size_t i = 0; i < postorder->size; i++) {
        if (wm_is_treenode_keyroot(postorder->nodes[i], postorder))
            wm_uintarray_push(ret, i);
    }

    wm_nodearray_free(postorder);
    return ret;
}

// https://github.com/timtadh/zhang-shasha
void wm_treedist(TreeNode *tree1, TreeNode *tree2, size_t i, size_t j,
                 UIntArray *dists, TreeEditDistanceCosts costs)
{
    assert(tree1);
    assert(tree2);

    NodeArray *tree1_postorder = wm_postorder_traversal(tree1);
    NodeArray *tree2_postorder = wm_postorder_traversal(tree2);

    UIntArray *tree1_lmds = wm_treenode_all_lmds_index(tree1);
    UIntArray *tree2_lmds = wm_treenode_all_lmds_index(tree2);

    size_t m = i - wm_uintarray_at(tree1_lmds, i) + 2;
    size_t n = j - wm_uintarray_at(tree2_lmds, j) + 2;

    UIntArray *fd = wm_uintarray_new();

    _Static_assert(sizeof(uint64_t) == sizeof(uint64_t*), "static assert failed");

    for (size_t k = 0; k < m; k++) {
        wm_uintarray_push(fd, (uint64_t)calloc(n, sizeof(uint64_t)));
    }

    size_t ioff = wm_uintarray_at(tree1_lmds, i) - 1;
    size_t joff = wm_uintarray_at(tree2_lmds, j) - 1;

    for (size_t x = 1; x < m; x++) {
        TreeNode *node = tree1_postorder->nodes[x + ioff];
        ((uint64_t**)fd->elements)[x][0] = ((size_t)wm_uintarray_2d_index(fd,
            x - 1, 0) + (size_t)costs.remove_cost);
    }

    for (size_t y = 1; y < n; y++) {
        TreeNode *node = tree2_postorder->nodes[y + joff];
        ((uint64_t**)fd->elements)[0][y] = ((size_t)wm_uintarray_2d_index(fd,
            0, y - 1) + (size_t)costs.insert_cost);
    }

    for (size_t x = 1; x < m; x++) {
        for (size_t y = 1; y < n; y++) {
            TreeNode *node1 = tree1_postorder->nodes[x + ioff];
            TreeNode *node2 = tree2_postorder->nodes[y + joff];
            if (wm_uintarray_at(tree1_lmds, i) == tree1_lmds->elements[x + ioff] &&
                wm_uintarray_at(tree2_lmds, j) == tree2_lmds->elements[y + joff]) {

                size_t _costs[3] = {
                    wm_uintarray_2d_index(fd, x - 1, y) + costs.remove_cost, //(node1)
                    wm_uintarray_2d_index(fd, x, y - 1) + costs.insert_cost, //(node2)
                    wm_uintarray_2d_index(fd, x - 1, y - 1) + (*costs.update_cost_function)(node1, node2),
                };

                size_t min = SIZE_MAX;
                for (size_t k = 0; k < 3; k++) {
                    if (_costs[k] < min) {
                        min = _costs[k];
                    }
                }

                assert(min != SIZE_MAX);

                ((uint64_t**)fd->elements)[x][y] = min;
                ((uint64_t**)dists->elements)[x + ioff][y + joff] = min;
            } else {
                size_t p = tree1_lmds->elements[x + ioff] - 1 - ioff;
                size_t q = tree2_lmds->elements[y + joff] - 1 - joff;

                size_t _costs[3] = {
                    (size_t)wm_uintarray_2d_index(fd, x - 1, y) + costs.remove_cost, // (node1)
                    (size_t)wm_uintarray_2d_index(fd, x, y - 1) + costs.insert_cost, // (node2)
                    (size_t)wm_uintarray_2d_index(fd, p, q) +
                    (size_t)wm_uintarray_2d_index(dists, x + ioff, y + joff),
                };

                size_t min = SIZE_MAX;
                for (size_t k = 0; k < 3; k++) {
                    if (_costs[k] < min) {
                        min = _costs[k];
                    }
                }

                assert(min != SIZE_MAX);
                ((uint64_t**)fd->elements)[x][y] = min;
            }
        }
    }

    for (size_t k = 0; k < m; k++) {
        free((void*)wm_uintarray_at(fd, k));
    }

    wm_nodearray_free(tree1_postorder);
    wm_nodearray_free(tree2_postorder);
    wm_uintarray_free(tree1_lmds);
    wm_uintarray_free(tree2_lmds);
    wm_uintarray_free(fd);
}

// https://github.com/timtadh/zhang-shasha
size_t wm_tree_edit_distance(TreeNode *tree1, TreeNode *tree2, TreeEditDistanceCosts costs)
{
    assert(tree1);
    assert(tree2);

    NodeArray *tree1_postorder = wm_postorder_traversal(tree1);
    NodeArray *tree2_postorder = wm_postorder_traversal(tree2);

    UIntArray *tree1_keyroot_indexes = wm_treenode_all_keyroots_index(tree1);
    UIntArray *tree2_keyroot_indexes = wm_treenode_all_keyroots_index(tree2);

    UIntArray *dists = wm_uintarray_new();

    _Static_assert(sizeof(uint64_t) == sizeof(uint64_t*), "static assert failed");

    for (size_t k = 0; k < tree1_postorder->size; k++) {
        wm_uintarray_push(dists, (uint64_t)calloc(tree2_postorder->size, sizeof(uint64_t)));
    }

    for (size_t p = 0; p < tree1_keyroot_indexes->size; p++) {
        for (size_t q = 0; q < tree2_keyroot_indexes->size; q++) {
            size_t i = wm_uintarray_at(tree1_keyroot_indexes, p);
            size_t j = wm_uintarray_at(tree2_keyroot_indexes, q);
            wm_treedist(tree1, tree2, i, j, dists, costs);
        }
    }

    size_t ret = wm_uintarray_2d_index(dists, tree1_postorder->size - 1, tree2_postorder->size - 1);

    for (size_t k = 0; k < tree1_postorder->size; k++) {
        free((void*)wm_uintarray_at(dists, k));
    }

    wm_nodearray_free(tree1_postorder);
    wm_nodearray_free(tree2_postorder);
    wm_uintarray_free(tree1_keyroot_indexes);
    wm_uintarray_free(tree2_keyroot_indexes);
    wm_uintarray_free(dists);

    return ret;
}

void wm_logentries_calculate_distances(UIntArray *log_entries, TreeNode *curr_tree,
                                       UpdateCostFunction update_cost_function)
{
    TreeEditDistanceCosts costs = (TreeEditDistanceCosts) {
        .insert_cost = 3,
        .remove_cost = 100,
        .update_cost_function = update_cost_function,
    };

    for (size_t i = 0; i < log_entries->size; i++) {
        LogEntry *entry = (LogEntry*)wm_uintarray_at(log_entries, i);

        for (size_t j = 0; j < entry->workspaces->size; j++) {
            TreeNode *node = entry->workspaces->nodes[j];
            if (node == NULL) continue;

            node->distance = wm_tree_edit_distance(curr_tree, node, costs);
        }
    }
}

void wm_treenode_print(TreeNode *node)
{
    char *str = wm_treenode_to_str(node);
    DEBUG_PRINT("%s\n", str);
    free(str);
}

TreeNode* wm_json_obj_to_treenode(struct json_object *jobj)
{
    assert(jobj);

    TreeNode *tree_ret = wm_treenode_new(NODE_VERTICAL, NULL);

    bool found_key = false;

    struct json_object *ret;
    struct json_object *tmp;

    char *key = "type";
    // TODO
    found_key = json_object_object_get_ex(jobj, key, &ret);
    if (!found_key) goto not_found_key;
    const char *type = json_object_get_string(ret);
    tree_ret->type = wm_node_type_from_str((char*)type, strlen(type));

    key = "parent";
    found_key = json_object_object_get_ex(jobj, key, &ret);
    if (!found_key) goto not_found_key;
    tree_ret->parent = (void*)json_object_get_int64(ret);

    key = "id";
    found_key = json_object_object_get_ex(jobj, key, &ret);
    if (!found_key) goto not_found_key;
    tree_ret->id = json_object_get_int64(ret);

    if (tree_ret->type == NODE_CLIENT) {
        tree_ret->client = calloc(1, sizeof(Client));
        key = "client";
        found_key = json_object_object_get_ex(jobj, key, &tmp);
        if (!found_key) goto not_found_key;

        key = "name";
        found_key = json_object_object_get_ex(tmp, key, &ret);
        if (!found_key) goto not_found_key;
        tree_ret->client->name = strdup(json_object_get_string(ret));

        key = "is_pid_set";
        found_key = json_object_object_get_ex(tmp, key, &ret);
        if (!found_key) goto not_found_key;
        tree_ret->client->is_pid_set = json_object_get_boolean(ret);

        key = "pid";
        found_key = json_object_object_get_ex(tmp, key, &ret);
        if (!found_key) goto not_found_key;
        tree_ret->client->pid = json_object_get_int(ret);
    }

    // pos
    key = "pos";
    found_key = json_object_object_get_ex(jobj, key, &tmp);
    if (!found_key) goto not_found_key;

    key = "x";
    found_key = json_object_object_get_ex(tmp, key, &ret);
    if (!found_key) goto not_found_key;
    tree_ret->pos.x = json_object_get_int(ret);

    key = "y";
    found_key = json_object_object_get_ex(tmp, key, &ret);
    if (!found_key) goto not_found_key;
    tree_ret->pos.y = json_object_get_int(ret);

    key = "w";
    found_key = json_object_object_get_ex(tmp, key, &ret);
    if (!found_key) goto not_found_key;
    tree_ret->pos.w = json_object_get_int(ret);

    key = "h";
    found_key = json_object_object_get_ex(tmp, key, &ret);
    if (!found_key) goto not_found_key;
    tree_ret->pos.h = json_object_get_int(ret);

    // children
    key = "children";
    found_key = json_object_object_get_ex(jobj, key, &ret);
    if (!found_key) goto not_found_key;
    struct array_list *list = json_object_get_array(ret);

    for (size_t i = 0; i < array_list_length(list); i++) {
        struct json_object *child = array_list_get_idx(list, i);
        wm_nodearray_push(tree_ret->children, wm_json_obj_to_treenode(child));
    }

    return tree_ret;

not_found_key:
    fprintf(stderr, "wm: %s: could not find key %s\n", __func__, key);
    return tree_ret;
}

#define CREATE_LOGFILE_ERROR(str) \
    do { \
        fprintf(stderr, "wm: could not create log file. error: %s. exiting.\n", str); \
        exit(1); \
    } while (0); \

static void recursive_mkdir(char *path, mode_t mode)
{
    char *sep = strrchr(path, '/');

    if(sep != NULL) {
        *sep = '\0';
        recursive_mkdir(path, mode);
        *sep = '/';
    }

    if(mkdir(path, mode) < 0 && errno != EEXIST) {
        fprintf(stderr, "wm: could not create directory. error %s. exiting.\n",
                strerror(errno));
        exit(1);
    }
}

void wm_logfile_init(const char *path)
{

    if (access(path, F_OK) == 0) return;

    char *dup = strdup(path);
    char *log_file_dir = dirname(dup);

    int ret = access(log_file_dir, W_OK | R_OK);

    if (ret == 0) {
        ret = creat(path, 0664);
        if (ret < 0) CREATE_LOGFILE_ERROR(strerror(errno));

        ret = close(ret);
        if (ret < 0) CREATE_LOGFILE_ERROR(strerror(errno));

        free(dup);
        return;
    }

    if (ret < 0 && errno == ENOENT) {
        recursive_mkdir(log_file_dir, 0751);

        ret = creat(path, 0664);
        if (ret < 0) CREATE_LOGFILE_ERROR(strerror(errno));

        ret = close(ret);
        if (ret < 0) CREATE_LOGFILE_ERROR(strerror(errno));
    } else {
        CREATE_LOGFILE_ERROR(strerror(errno));
    }

    free(dup);
}

/* UIntArray<LogEntry*> */
UIntArray* wm_read_log(const char* filename)
{
    _Static_assert(sizeof(LogEntry*) == sizeof(uint64_t), "");
    // TODO maybe prealloc
    UIntArray* ret = wm_uintarray_new();

    int fd = -1;

    struct json_object *jobj = NULL;

    fd = open(filename, O_RDONLY);

    if (fd < 0) {
        perror("open");
        return ret;
    }

    jobj = json_object_from_fd(fd);
    if (!jobj) {
        const char *err = json_util_get_last_err();
        fprintf(stderr, "wm: could not read json: %s\n", err);
        return ret;
    }

    char *token;
    const char *name = NULL;
    struct json_object *value = NULL;
    struct json_object_iterator iter = json_object_iter_begin(jobj);
    struct json_object_iterator iter_end = json_object_iter_end(jobj);

    const char *ws_name = NULL;
    struct json_object *ws_value = NULL;
    struct json_object_iterator ws_iter = json_object_iter_init_default();
    struct json_object_iterator ws_iter_end = json_object_iter_init_default();

    while (!json_object_iter_equal(&iter, &iter_end)) {
        // TODO maybe prealloc
        LogEntry *entry = calloc(1, sizeof(LogEntry));
        entry->workspaces = wm_nodearray_new();
        name = json_object_iter_peek_name(&iter);
        value = json_object_iter_peek_value(&iter);

        int i = 0;
        char *token = strtok((char*)name, ":");

        while(token != NULL) {
            i++;

            switch (i) {
            case 1:
                // TODO
                entry->timestamp = atol(token);
                break;
            case 2:
                strncpy(entry->function_name, token, 128);
                break;
            case 3:
                if (strncmp(token, "after", strlen("after")) == 0)
                    entry->after = true;
                else {
                    fprintf(stderr, "wm: encountered invalid token %s\n. exiting.",
                            token);
                    abort();
                }
                break;
            default:
                fprintf(stderr, "wm: encountered invalid token %s\n. exiting.",
                        token);
                abort();
            }

            token = strtok(NULL, ":");
        }

        ws_iter = json_object_iter_begin(value);
        ws_iter_end = json_object_iter_end(value);

        while (!json_object_iter_equal(&ws_iter, &ws_iter_end)) {
            ws_name = json_object_iter_peek_name(&ws_iter);
            ws_value = json_object_iter_peek_value(&ws_iter);

            int ws_index = atoi(ws_name);

            while (entry->workspaces->size < ws_index) {
                wm_nodearray_push(entry->workspaces, NULL);
            }

            wm_nodearray_push(entry->workspaces, wm_json_obj_to_treenode(ws_value));

            json_object_iter_next(&ws_iter);
        }

        wm_uintarray_push(ret, (uint64_t)entry);
        json_object_iter_next(&iter);
    }

    json_object_put(jobj);

    int result = close(fd);
    if (result < 0)
        perror("close");

    return ret;
}

void wm_tree_to_DOT(TreeNode *root, const char *filename)
{
    DEBUG_PRINT("%s\n", __func__);

    FILE *fp = fopen(filename, "w");
    assert(fp);

    fprintf(fp, "digraph {\n");

    NodeArray *all_nodes = wm_all_nodes_to_array(root);
    NodeArray *printed_nodes = wm_nodearray_new();

    const int bufsize = 100;
    char buf1[bufsize];
    char buf2[bufsize];

    for (size_t i = 0; i < all_nodes->size; i++) {
        TreeNode *node = all_nodes->nodes[i];

        bool contains = false;
        for (size_t j = 0; j < printed_nodes->size; j++) {
            if (printed_nodes->nodes[j]->id == node->id)
                contains = true;
        }

        if (contains)
            continue;

        for (size_t j = 0; j < node->children->size; j++) {
            memset(buf1, 0, bufsize);
            memset(buf2, 0, bufsize);

            wm_node_type_to_str(node->type, buf1, bufsize);
            wm_node_type_to_str(node->children->nodes[j]->type, buf2, bufsize);

            fprintf(fp, "_%d_%s -> _%d_%s;\n", node->id, buf1,
                    node->children->nodes[j]->id, buf2);
        }
        wm_nodearray_push(printed_nodes, node);
    }

    fprintf(fp, "}");

    fflush(fp);
    fclose(fp);

    wm_nodearray_free(printed_nodes);
    wm_nodearray_free(all_nodes);
}

void wm_treenode_remove_node(Wm *wm, TreeNode *root, TreeNode *node)
{
    DEBUG_PRINT("%s\n", __func__);

    wm_tree_to_DOT(root, "wm_output.dot");

    assert(node);

    TreeNode *parent = node->parent;

    if (node->parent == NULL) {
        node->client = NULL;
        return;
    }

    for (size_t i = 0; i < parent->children->size; i++) {
        if (parent->children->nodes[i]->type == NODE_CLIENT &&
            parent->children->nodes[i] != node) {

            parent->type = NODE_CLIENT;
            parent->client = parent->children->nodes[i]->client;
            wm_treenode_free(node);
            wm_treenode_free(parent->children->nodes[i]);
            wm_nodearray_clear(parent->children);
            return;
        }
    }

    DEBUG_PRINT("%s: other_client was NULL!\n", __func__);

    TreeNode *sibling_node = wm_treenode_split_get_sibling(node);

    // wm_nodearray_free(&node->children);

    // TODO
    size_t dock_y = 0;

    if (parent->parent == NULL) {
        // parent is root node
        DEBUG_PRINT("parent is root node!\n");
        assert(node->type == NODE_CLIENT);

        TreeNode old_root = *node->client->ws->tree;

        sibling_node->parent = NULL;
        *node->client->ws->tree = *sibling_node;

        wm_nodearray_free(old_root.children);
        free(sibling_node);

        node->client->ws->tree->pos = (Rect) {
            .x = wm->config.border_width, .y = wm->config.border_width + dock_y ,
            .w = node->client->ws->monitor->info.width - wm->config.border_width*2,
            .h = node->client->ws->monitor->info.height - wm->config.border_width*2 - dock_y,
        };

        assert(node->client->ws->tree->children->size == 2);
        wm_treenode_split_space(node->client->ws->tree, &node->client->ws->tree->children->nodes[0]->pos,
                                &node->client->ws->tree->children->nodes[1]->pos);
    } else {
        sibling_node->parent = parent->parent;
        *parent = *sibling_node;

        for (size_t i = 0; i < parent->children->size; i++) {
            parent->children->nodes[i]->parent = parent;
        }

        wm_treenode_recalculate_space(parent->parent);
        wm_treenode_recalculate_space(parent);
    }

    assert(node->client->ws->tree->children->size >= 1);

    wm_treenode_free(node);
}

TreeNode* wm_treenode_split_get_sibling(TreeNode *node)
{
    for (size_t i = 0; i < node->parent->children->size; i++) {
        TreeNode *sibling_node = node->parent->children->nodes[i];
        if (sibling_node->id != node->id) {
            return sibling_node;
        }
    }

    return NULL;
}

TreeNode* wm_treenode_remove_client(Wm *wm, TreeNode *root, Client *client)
{
    DEBUG_PRINT("%s\n", __func__);
    TreeNode *client_node = wm_treenode_ptr_find_client_node(root, client);

    wm_tree_to_DOT(root, "wm_output.dot");

    if (client_node->parent == NULL) {
        client_node->client = NULL;
        return NULL;
    }

    TreeNode *node = client_node->parent;

    for (size_t i = 0; i < node->children->size; i++) {
        if (node->children->nodes[i]->type == NODE_CLIENT &&
            node->children->nodes[i]->client != client) {

            node->type = NODE_CLIENT;
            node->client = node->children->nodes[i]->client;
            wm_treenode_free(client_node);
            wm_treenode_free(node->children->nodes[i]);
            wm_nodearray_clear(node->children);
            return node;
        }
    }

    DEBUG_PRINT("%s: other_client was NULL!\n", __func__);

    TreeNode *sibling_node = wm_treenode_split_get_sibling(client_node);
    wm_treenode_free(client_node);

    // wm_nodearray_free(&node->children);

    // TODO
    size_t dock_y = 0;

    if (node->parent == NULL) {
        // parent is root node
        DEBUG_PRINT("parent is root node!\n");

        TreeNode old_root = *client->ws->tree;

        sibling_node->parent = NULL;
        *client->ws->tree = *sibling_node;

        wm_nodearray_free(old_root.children);
        free(sibling_node);

        client->ws->tree->pos = (Rect) {
            .x = wm->config.border_width, .y = wm->config.border_width + dock_y ,
            .w = client->ws->monitor->info.width - wm->config.border_width*2,
            .h = client->ws->monitor->info.height - wm->config.border_width*2 - dock_y,
        };

        assert(client->ws->tree->children->size == 2);

        client->ws->tree->children->nodes[0]->parent = client->ws->tree;
        client->ws->tree->children->nodes[1]->parent = client->ws->tree;
        wm_treenode_split_space(client->ws->tree, &client->ws->tree->children->nodes[0]->pos,
                                &client->ws->tree->children->nodes[1]->pos);
    } else {
        sibling_node->parent = node->parent;
        *node = *sibling_node;

        for (size_t i = 0; i < node->children->size; i++) {
            node->children->nodes[i]->parent = node;
        }

        wm_treenode_recalculate_space(node->parent);
        wm_treenode_recalculate_space(node);
    }

    assert(client->ws->tree->children->size >= 1);
    return client->ws->tree->children->nodes[0];
}

NodeArray* wm_nodearray_new()
{
    NodeArray *arr = calloc(1, sizeof(NodeArray));
    assert(arr);

    arr->capacity = 10;
    arr->nodes = calloc(arr->capacity, sizeof(TreeNode*));
    assert(arr->nodes);
    arr->size = 0;

    return arr;
}

void wm_nodearray_push(NodeArray *arr, TreeNode *node)
{
    assert(arr);

    arr->size++;

    if (arr->size >= arr->capacity) {
        TreeNode** temp = calloc(arr->capacity, sizeof(TreeNode*));
        assert(temp);
        memcpy(temp, arr->nodes, arr->capacity * sizeof(TreeNode*));
        free(arr->nodes);
        size_t old_capacity = arr->capacity;
        arr->capacity = old_capacity * 2;

        arr->nodes = calloc(arr->capacity, sizeof(TreeNode*));
        assert(arr->nodes);
        memcpy(arr->nodes, temp, old_capacity * sizeof(TreeNode*));

        free(temp);
        arr->nodes[arr->size - 1] = node;
        return;
    }

    arr->nodes[arr->size - 1] = node;
}

bool wm_nodearray_pop(NodeArray *arr, TreeNode **ret)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(arr);
    assert(ret);

    if (arr->size == 0)
        return false;

    *ret = arr->nodes[arr->size - 1];
    arr->size--;
    return true;
}

bool wm_nodearray_pop_front(NodeArray *arr, TreeNode **ret)
{
    assert(arr);
    assert(ret);

    if (arr->size == 0)
        return false;

    *ret = arr->nodes[0];
    arr->size--;
    memmove(arr->nodes, arr->nodes+1, arr->size * sizeof(TreeNode*));

    return true;
}

void wm_nodearray_clear(NodeArray *arr)
{
    DEBUG_PRINT("%s\n", __func__);
    arr->size = 0;
    memset(arr->nodes, 0, arr->capacity * sizeof(TreeNode*));
}

bool wm_nodearray_remove(NodeArray *arr, size_t index)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(arr);

    if (!arr)
        return false;

    if (index >= arr->size)
        return false;

    if (index == arr->size - 1) {
        arr->size--;
        return true;
    }

    memmove(arr->nodes + index, arr->nodes + index+1, arr->size-1 * sizeof(TreeNode*));

    return true;
}

void wm_nodearray_free(NodeArray *arr)
{
    assert(arr);

    arr->capacity = 0;
    arr->size = 0;
    free(arr->nodes);
    free(arr);
}

UIntArray* wm_uintarray_new()
{
    UIntArray *arr = calloc(1, sizeof(UIntArray));
    assert(arr);

    arr->capacity = 10;
    arr->elements = calloc(arr->capacity, sizeof(void*));
    assert(arr->elements);
    arr->size = 0;

    return arr;
}

uint64_t wm_uintarray_at(UIntArray *arr, size_t index)
{
    assert(arr);
    assert(index < arr->size);

    return arr->elements[index];
}

void wm_uintarray_push(UIntArray *arr, uint64_t element)
{
    assert(arr);

    arr->size++;

    if (arr->size >= arr->capacity) {
        void* temp = calloc(arr->capacity, sizeof(uint64_t));
        assert(temp);
        memcpy(temp, arr->elements, arr->capacity * sizeof(uint64_t));
        free(arr->elements);
        size_t old_capacity = arr->capacity;
        arr->capacity = old_capacity * 2;

        arr->elements = calloc(arr->capacity, sizeof(uint64_t));
        assert(arr->elements);
        memcpy(arr->elements, temp, old_capacity * sizeof(uint64_t));

        free(temp);
        arr->elements[arr->size - 1] = element;
        return;
    }

    arr->elements[arr->size - 1] = element;
}

bool wm_uintarray_pop(UIntArray *arr, uint64_t *ret)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(arr);
    assert(ret);

    if (arr->size == 0)
        return false;

    *ret = arr->elements[arr->size - 1];
    arr->size--;
    return true;
}

bool wm_uintarray_pop_front(UIntArray *arr, uint64_t *ret)
{
    assert(arr);
    assert(ret);

    if (arr->size == 0)
        return false;

    *ret = arr->elements[0];
    arr->size--;
    memmove(arr->elements, arr->elements+1, arr->size * sizeof(uint64_t));

    return true;
}

bool wm_uintarray_remove(UIntArray *arr, size_t index)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(arr);

    if (!arr)
        return false;

    if (index >= arr->size)
        return false;

    if (index == arr->size - 1) {
        arr->size--;
        return true;
    }

    memmove(arr->elements + index, arr->elements + index+1, arr->size-1 * sizeof(uint64_t));

    return true;
}

void wm_uintarray_free(UIntArray *arr)
{
    // DEBUG_PRINT("%s\n", __func__);
    assert(arr);
    RETURN_IF_NULL(arr->elements);

    arr->capacity = 0;
    arr->size = 0;
    free(arr->elements);
    arr->elements = NULL;

    free(arr);
}

uint64_t wm_uintarray_2d_index(UIntArray *arr, size_t i, size_t j)
{
    assert(arr);

    return ((uint64_t**)arr->elements)[i][j];
}

void wm_logentries_free(UIntArray *entries)
{
    for (size_t i = 0; i < entries->size; i++) {
        LogEntry *entry = (LogEntry*)wm_uintarray_at(entries, i);

        for (size_t j = 0; j < entry->workspaces->size; j++) {
            TreeNode *node = entry->workspaces->nodes[j];
            if (node == NULL) continue;

            NodeArray *children = wm_postorder_traversal(node);
            for (size_t k = 0; k < children->size; k++) {
                if (children->nodes[k]->type == NODE_CLIENT) {
                    free(children->nodes[k]->client->name);
                    free(children->nodes[k]->client);
                }
                wm_treenode_free(children->nodes[k]);
            }

            wm_nodearray_free(children);
        }

        wm_nodearray_free(entry->workspaces);
        free(entry);
    }

    wm_uintarray_free(entries);
}

TreeNode* wm_treenode_ptr_find_focused_client_node(TreeNode *root)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(root);

    TreeNode *ret = NULL;

    NodeArray *visited = wm_nodearray_new();
    wm_nodearray_push(visited, root);

    NodeArray *queue = wm_nodearray_new();

    wm_nodearray_push(queue, root);

    while (queue->size > 0) {
        TreeNode *node;

        if (!wm_nodearray_pop_front(queue, &node))
            goto ret;

        if (node->type == NODE_CLIENT && node->client && node->client->focused) {
            ret = node;
            goto ret;
        }

        for (size_t i = 0; i < node->children->size; i++) {
            TreeNode *child_node = node->children->nodes[i];
            bool _visited = false;

            for (size_t j = 0; j < visited->size; j++) {
                if (visited->nodes[j]->id == child_node->id) {
                    _visited = true;
                    break;
                }
            }

            if (!_visited) {
                wm_nodearray_push(visited, child_node);
                wm_nodearray_push(queue, child_node);
            }
        }
    }

ret:
    wm_nodearray_free(visited);
    wm_nodearray_free(queue);
    return ret;
}

TreeNode* wm_treenode_ptr_find_client_node(TreeNode *root, Client *client)
{
    assert(root);
    assert(client);

    TreeNode *ret = NULL;
    NodeArray *client_nodes = wm_treenode_find_client_nodes(root);

    for (size_t i = 0; i < client_nodes->size; i++) {
        TreeNode *node = (TreeNode*)client_nodes->nodes[i];
        if (node->client == client) {
            ret = node;
            goto ret;
        }
    }

ret:
    wm_nodearray_free(client_nodes);
    return ret;
}

NodeArray *wm_treenode_find_client_nodes(TreeNode *root)
{
    assert(root);

    NodeArray *ret = wm_nodearray_new();
    NodeArray *all_nodes = wm_all_nodes_to_array(root);

    for (size_t i = 0; i < all_nodes->size; i++) {
        TreeNode *node = all_nodes->nodes[i];
        if (node->type == NODE_CLIENT && node->client)
            wm_nodearray_push(ret, node);
    }

    wm_nodearray_free(all_nodes);
    return ret;
}

NodeArray* wm_all_nodes_to_array(TreeNode *root)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(root);

    NodeArray *ret = wm_nodearray_new();

    NodeArray *visited = wm_nodearray_new();
    wm_nodearray_push(visited, root);

    NodeArray *queue = wm_nodearray_new();

    wm_nodearray_push(queue, root);

    while (queue->size > 0) {
        TreeNode *node;

        if (!wm_nodearray_pop_front(queue, &node))
            goto ret;

        wm_nodearray_push(ret, node);

        for (size_t i = 0; i < node->children->size; i++) {
            TreeNode *child_node = node->children->nodes[i];
            bool _visited = false;

            for (size_t j = 0; j < visited->size; j++) {
                if (visited->nodes[j]->id == child_node->id) {
                    _visited = true;
                    break;
                }
            }

            if (!_visited) {
                wm_nodearray_push(visited, child_node);
                wm_nodearray_push(queue, child_node);
            }
        }
    }

ret:
    wm_nodearray_free(visited);
    wm_nodearray_free(queue);
    return ret;
}
